import os; os.system('clear')
from paho.mqtt import client as mqtt_client
import numpy as np
import ciervo.parameters as p
from ciervo.aux_tools import Buffer
import re
from time import sleep

def natural_sort(l): 
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
    return sorted(l, key=alphanum_key)


broker = '127.0.0.1'
port = 1883
topic = "data"


class StoreStream:
    def __init__(self, folder='recordings', duration=60*60*3):
        self.duration = duration  # in seconds
        self.num_points = self.duration * p.SAMPLE_RATE
        self.folder = folder

        # Buffer
        self.file_num = 0
        self.buffer = Buffer(self.duration)

        # MQTT
        self.client = mqtt_client.Client(mqtt_client.CallbackAPIVersion.VERSION2, 'saver')
        self.client.on_connect = on_connect
        self.client.on_message = self.on_message
        self.client.connect(broker, port)
        self.client.subscribe(topic)
        self.client.loop_start()

        sleep(1)

        # File name input
        print("Enter file name: ", end='')
        self.file_name = input()
        if '.npy' not in self.file_name:
            self.file_name += '.npy'


        while True:
            # Ask to stop the recording
            print("Press 'q' to stop recording: ", end='')
            command = input()
            if command == 'q':
                self.client.loop_stop()
                break
        
        self.save_data()



    def on_message(self, client, userdata, msg):
        data = np.frombuffer(msg.payload, dtype=p.PRECISION)
        data = data.reshape(p.NUM_CHANNELS, -1)
        self.buffer.data = data


    def save_data(self):
        filename = os.path.join(self.folder, f"{self.file_name}")
        np.save(filename, self.buffer.data[:, :self.buffer.idx])
        print(f"Data saved to {filename}")
        self.file_num += 1

# The callback function of connection
def on_connect(client, userdata, flags, rc, properties):
    print(f"Connected with result code {rc}")
    client.subscribe('data')


def verify_recording(folder):
    files = os.listdir(folder)
    files = natural_sort([f for f in files if f.endswith('.npy')])
    data = []
    for f in files:
        data.append(np.load(os.path.join(folder, f)))
    data = np.concatenate(data, axis=1)

    channels, points = data.shape
    print(f"Data shape: {data.shape}")
    print(f"Number of channels: {channels}")
    print(f"Number of points: {points}")
    print(f"Duration: {points / p.SAMPLE_RATE} seconds")

    print(set(np.diff(data[0])))




if __name__ == '__main__':
    StoreStream()
    verify_recording('recordings')




import numpy as np
from scipy import signal
import scipy.stats

# Optimized function for feature extraction
def features_v1(data, divide=3, fs=250):
    # data : (C, T) 
    # T: number of samples
    # C: number of channels
    # divide: number of segments to divide the signal

    C, T = data.shape
    result = []
    feature_names = []
    
    segment_size = T // divide  # Calculate segment size once

    for c in range(C):
        signal0 = data[c, :]

        # Full wave rectification
        rectified_signal = np.abs(signal0)

        # Hilbert envelope (compute for each channel only once)
        env = np.abs(signal.hilbert(signal0))

        # RMS
        rms = np.sqrt(np.mean(rectified_signal**2))
        result.append(rms)
        feature_names.append(f"rms_channel_{c}")

        # Variance
        var = np.var(rectified_signal)
        result.append(var)
        feature_names.append(f"var_channel_{c}")

        # Kurtosis and skewness
        kurt = scipy.stats.kurtosis(rectified_signal)
        skew = scipy.stats.skew(rectified_signal)
        result.extend([kurt, skew])
        feature_names.extend([f"kurt_channel_{c}", f"skew_channel_{c}"])

        # Zero crossings
        zc = np.sum(np.diff(np.sign(signal0)) != 0)
        result.append(zc)
        feature_names.append(f"zc_channel_{c}")

        # Frequency domain features
        feature_names.extend([f"median_freq_channel_{c}", f"mean_freq_channel_{c}", f"peak_freq_channel_{c}"])


        mean_env = np.mean(env)
        std_env = np.std(env)
        max_env = np.max(env)
        min_env = np.min(env)

        result.extend([mean_env, std_env, max_env, min_env])

    return np.array(result), feature_names




import torch 

def get_device():
    if torch.cuda.is_available():
        device = torch.device('cuda')
        print("Using CUDA (GPU)")
    elif torch.backends.mps.is_available():
        device = torch.device('mps')
        print("Using MPS (Apple Silicon GPU)")
    else:
        device = torch.device('cpu')
        print("Using CPU")
    
    return device


import lightning as L
import torch.nn as nn
import torch
import torch.nn.functional as F


class GaitRNN(L.LightningModule):
    def __init__(self, n_classes, n_features, n_hidden, n_layers, num_channels):
        super().__init__()

        # Model
        self.linear = nn.Linear(num_channels, 1)
        self.lstm = nn.LSTM(n_features, n_hidden, n_layers, batch_first=True)
        self.fc = nn.Linear(n_hidden, n_classes)

    def forward(self, x):
        # N ~ batch size
        # C ~ number of channels
        # F ~ number of features
        # T ~ sequence length

        #  x ~ (N, C, F, T )
        x = x.permute(0, 3, 2, 1) # (N, T, F, C)
        x = self.linear(x) # (N, T, F, 1)
        x = x.squeeze(3) # (N, T, F)

        # rnn
        out, _ = self.lstm(x)  # (N, T, n_hidden)
        out = self.fc(out[:, -1, :] ) 
        return out

    def training_step(self, batch, batch_idx):
        x, y_target = batch
        
        # forward
        y_hat = self(x)

        # loss
        loss = F.cross_entropy(y_hat, y_target)

        # log
        self.log('loss/train', loss, prog_bar=True)
        acc = (y_hat.argmax(1) == y_target).float().mean()
        self.log('acc/train', acc, prog_bar=True)

        return loss
    
    def validation_step(self, batch, batch_idx):
        x, y_target = batch

        # forward
        y_hat = self(x)

        # loss
        loss = F.cross_entropy(y_hat, y_target)

        # log
        self.log('loss/val', loss, prog_bar=True)
        acc = (y_hat.argmax(1) == y_target).float().mean()
        self.log('acc/val', acc, prog_bar=True)

        return loss
    
    def test_step(self, batch, batch_idx):
        x, y_target = batch

        # forward
        y_hat = self(x)
        loss = F.cross_entropy(y_hat, y_target)

        # log
        self.log('loss/test', loss)
        acc = (y_hat.argmax(1) == y_target).float().mean()
        self.log('acc/test', acc)

        return loss

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters(), lr=0.01)


import os; os.system('clear')
from paho.mqtt import client as mqtt_client
from ciervo.models import features_v1
import numpy as np
import ciervo.parameters as p
from ciervo.aux_tools import Buffer
import time
import joblib
import serial
import struct
from threading import Thread


broker = p.BROKER_HOST
port = 1883
topic = "data"


class SendAngleSerial:

    def __init__(self):
        self.ser = serial.Serial('/dev/ttyACM0',
                                 baudrate=9600,
                                 bytesize=serial.EIGHTBITS,
                                 parity=serial.PARITY_NONE,
                                 stopbits=serial.STOPBITS_ONE,
                                 timeout=1)
        
        self.msg_tx = 0
        self.msg_rx = 0

        time.sleep(2)

        frequency = 10 # Hz

        thread = Thread(target = self.read_byte_loop, args=())
        thread.start()

    def send_byte(self, value):
        # Send int
        int_value = int(value)

        if 0 <= int_value <= 255:
            pass
        else:
            if int_value > 255:
                int_value = 255
            
            elif int_value < 0:
                int_value = 0

        self.ser.write(bytes([int_value]))

        self.msg_tx = int_value

        print(f'raw_msg:\t{value}\t,\tmsg_tx:\t{self.msg_tx}\t,\tmsg_rx:\t{self.msg_rx}')
    
    def read_byte(self):
        if self.ser.in_waiting > 0:
            received_data = self.ser.readline()
            #received_value = int.from_bytes(received_data, byteorder='big')
            decoded_data = received_data.decode('utf-8').strip()
            int_val = int(decoded_data)

            self.msg_rx = int_val

            return received_data
    
    def read_byte_loop(self):
        try:
            while True:
                self.read_byte()

        except:
            return



class RealTimeInference:
    def __init__(self, 
                 window=0.1, 
                 emg_prepro=None, 
                 emg_model=None, 
                 emg_idx=[0, 1, 2, 3],
                 acc_idx=[8, 9, 10],
                 serial_send=True,
                 ):
        self.update_speed = 1/10 # seconds
        self.window = window  # seconds
        self.angle_speed = 5 # degrees per second
        self.serial_send = serial_send

        self.emg_idx = emg_idx
        self.acc_idx = acc_idx

        if self.serial_send:
            self.serial = SendAngleSerial()

        # Angle
        self._angle = 0

        # preprocesamiento
        self.emg_prepro = emg_prepro

        # modelo
        self.emg_model = emg_model
        if self.emg_model is not None:
            self.emg_model = joblib.load(emg_model)

        # Buffer
        self.buffer = Buffer(self.window, roll=True)

        # pre-compile emg_prepro
        if self.emg_prepro is not None:
            self.emg_prepro(self.buffer.data[self.emg_idx, :])


        # MQTT
        self.client = mqtt_client.Client(mqtt_client.CallbackAPIVersion.VERSION2, 'inference')
        self.client.on_connect = on_connect
        self.client.on_message = self.on_message
        self.client.connect(broker, port)
        self.client.subscribe(topic)
        self.client.loop_start()

        self.update()

    @property
    def angle(self):
        return self._angle
    
    @angle.setter
    def angle(self, value):
        # min 90, max 180
        self._angle = np.clip(value, 90, 175)

    def on_message(self, client, userdata, msg):
        # Actualizar buffer
        data = np.frombuffer(msg.payload, dtype=p.PRECISION)
        data = data.reshape(p.NUM_CHANNELS, -1)
        self.buffer.data = data

    def update(self):
        while True:
            tic = time.time()
            data = self.buffer.data
            
            # data
            emg = data[self.emg_idx, :]
            #acc = data[self.acc_idx, :]
            if self.emg_prepro is not None:
                features, _ = self.emg_prepro(emg) # (C, T)
                # add dimension
                features = features.reshape(1, -1)
            
            if self.emg_model is not None:
    
                prediction = self.emg_model.predict(features)[0]
                #print("prediccion", prediction)
                # Update angle
                if prediction == 1: #activa
                    self.angle +=  10
                else: 
                    self.angle =  90

                # Send angle
                if self.serial_send:
                    self.serial.send_byte(self.angle)

                self.client.publish('marker', int(self._angle), qos=0)
                #print(self.angle)



            toc = time.time()
            #print((toc-tic)/self.update_speed)
            time.sleep(np.max([self.update_speed - (toc-tic), 0]))


 


# The callback function of connection
def on_connect(client, userdata, flags, rc, properties):
    print(f"Connected with result code {rc}")
    client.subscribe('data')


if __name__ == '__main__':

    RealTimeInference(
        emg_prepro=features_v1, 
        emg_model='model.pkl',
        serial_send=True,
        )



import os; os.system('clear')
import numpy as np
from scipy import signal
import scipy
from ciervo.models import train_test_split
import pandas as pd
import ciervo.parameters as p
from ciervo.models import features_v1
# train model
from sklearn.ensemble import RandomForestClassifier
# svm
from sklearn.svm import SVC
from sklearn.metrics import accuracy_score



if __name__ == "__main__":
    # cargar data
    data = np.load('recordings/otro6.npy')

    n_points = data.shape[1]
    print(f"Tiempo de grabación: {n_points/250} s")

    df = pd.DataFrame(data.T, columns=p.CH_NAMES)
    df['labels'] = df['MARKERS']
    # remove df['labels'] == 0
    df = df[df['labels'] != 0]
    df['labels'] = df['labels'].apply(lambda x: 1 if x == 2 else 0)

    train_data, train_label,  test_data, test_label = train_test_split([df], 
                                         columna=['C1', 'C2', 'C3', 'C4'],
                                         window_size=25,
                                         overlap=0, 
                                         test_size=0.2,
                                         random_state=42)
    print(train_data.shape, train_label.shape, test_data.shape, test_label.shape)

    # feature extraction
    train_features = []
    for i in range(len(train_data)):
        train_features.append(features_v1(train_data[i].T)[0])
    train_features = np.array(train_features)

    test_features = []
    for i in range(len(test_data)):
        test_features.append(features_v1(test_data[i].T)[0])
    test_features = np.array(test_features)

    print(train_features.shape, test_features.shape)
    print(train_label.shape, test_label.shape)


    clf = RandomForestClassifier()
    clf.fit(train_features, train_label)
    y_pred = clf.predict(test_features)

    print(accuracy_score(test_label, y_pred))
    
    # save model
    import joblib
    joblib.dump(clf, 'model.pkl')

    # load model
    clf = joblib.load('model.pkl')
    y_pred = clf.predict(test_features)
    print(accuracy_score(test_label, y_pred))








import numpy as np
import ciervo.parameters as p


class Buffer:
    def __init__(self, duration, roll=False):
        self.window = int(duration * p.SAMPLE_RATE)
        self._data = np.zeros((p.NUM_CHANNELS, self.window ), dtype=p.PRECISION)
        self.idx = 0
        self.roll = roll

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        if self.roll:
            self._data = np.roll(self._data, -data.shape[1], axis=1)
            self._data[:, -data.shape[1]:] = data

        else:
            self._data[:, self.idx:self.idx+data.shape[1]] = data
            self.idx += data.shape[1]



from torch.utils.data import Dataset
from matplotlib import pyplot as plt
import torchaudio
import torch


class SpectrogramDataset(Dataset):
    
    def __init__(self, data, label, n_fft=16):
        self.data = data
        self.label = label
        self.transform = torchaudio.transforms.Spectrogram(n_fft=n_fft)
        self.cache = {}

    def __len__(self):
        return len(self.label)
    
    def __getitem__(self, idx):
        data = torch.tensor(self.data[idx]).float()
        label = torch.tensor(self.label[idx]).type(torch.LongTensor)

        if idx not in self.cache:
            W, C = data.shape        
            # Convert to spectrogram
            spec_data = torch.stack([self.transform(data[:, c]) for c in range(C)])
            self.cache[idx] = spec_data
        else:
            spec_data = self.cache[idx]


        return spec_data, label
    
    def plot_sample(self):
        spec_data, _ = self.__getitem__(0)

        C = len(spec_data)
        fig, ax = plt.subplots(C, 1, figsize=(10, 10))
        for c in range(C):
            spec_db = torchaudio.transforms.AmplitudeToDB('power',top_db=80)(spec_data[c])
            ax[c].imshow(spec_db.numpy(), aspect='auto', origin='lower')
            
        plt.show()




import numpy as np

# MQTT params
BROKER_HOST = '127.0.0.1'
URL_INFLUXDB = 'http://100.90.57.1:8086'
BROKER_PORT = 1883
PRECISION = np.float32


PACKAGE_CHANNEL = 0
CH_NAMES = ['C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8', 'AX', 'AY', 'AZ', 'TIME', 'PKG', 'MARKERS']

# Real stream
EMG_CHANNELS = [1, 2, 3, 4, 5, 6, 7,8 ]  # 
ACC_CHANNELS = [9, 10, 11]
TIME_CHANNEL = 22
MARKER_CHANNEL = 23
NUM_CHANNELS  = len(EMG_CHANNELS + ACC_CHANNELS + [TIME_CHANNEL]+ [PACKAGE_CHANNEL]  + [MARKER_CHANNEL])
ALL_CHANNELS = EMG_CHANNELS + ACC_CHANNELS + [TIME_CHANNEL] + [PACKAGE_CHANNEL ] + [MARKER_CHANNEL]

# synthetic streamn
SYN_EMG_CHANNELS = [1, 2, 3, 4, 5, 6, 7, 8]  #
SYN_ACC_CHANNELS = [17, 18, 19]
SYN_TIME_CHANNEL = 30
SYN_MARKER_CHANNEL = 31
SYN_NUM_CHANNELS = len(SYN_EMG_CHANNELS + SYN_ACC_CHANNELS + [SYN_TIME_CHANNEL] + [PACKAGE_CHANNEL] + [SYN_MARKER_CHANNEL])
SYN_ALL_CHANNELS = SYN_EMG_CHANNELS + SYN_ACC_CHANNELS + [SYN_TIME_CHANNEL] + [PACKAGE_CHANNEL ] + [SYN_MARKER_CHANNEL]

SAMPLE_RATE = 250

# INFLUX
TOPIC = 'data'
TOKEN = ''
INFLUXDB_TOKEN= "holas"
INFLUXDB_ORG = "Ciervo"
INFLUXDB_BUCKET = "Ciervo"



from typing import Union
import pandas as pd
from scipy.signal import find_peaks
from matplotlib import pyplot as plt
from ciervo.procesamiento import ButterLowpassFilter
import numpy as np



# data is list or dataframe
def label_data(data: Union[list, pd.DataFrame], num_fases: int = 10) -> pd.DataFrame:
    """
    Labels the data with phases based on the peaks in the 'Angle' column.
    Adds a new column 'labels' to the data with the phase labels.
    Args:
        data (Union[list, pd.DataFrame]): The data to be labeled. It can be either a list of pandas DataFrames or a single DataFrame.
        num_fases (int, optional): The number of phases to label each step. Defaults to 10.
    Returns:
        pd.DataFrame: The labeled data.
    Raises:
        None
    """
    if isinstance(data, pd.DataFrame):
        data = [data]

    low_pass = ButterLowpassFilter(20) 

    df_list = []
    for df in data:
        labels = []

        # Filtrado de los angulos para evitar picos
        df.loc[:, 'Angle'] = low_pass.apply(df['Angle'])

        # Encontrar los picos
        peaks = find_peaks(df['Angle'], height=50, distance=125, width=50)[0]


        # Iterar en cada paso para etiquetar
        for i in range(0, len(peaks)-1):    
            # Etiquetado al comienzo siempre es -1
            if i == 0:
                labels.extend([-1] * peaks[0]) 

            # Etiquetado de cada paso en fases
            temp = np.array([-1]* (peaks[i+1] - peaks[i]))
            for phase in range(num_fases):
                temp[int(len(temp)/num_fases*phase):int(len(temp)/num_fases*(phase+1))] = float(phase)
            labels.extend(temp)

        # Etiquetado al final siempre es -1
        labels.extend([-1] * (len(df['Angle']) - len(labels)))   
                    
        df['labels'] = np.array(labels)

        # Remover -1
        df = df[df.labels != -1].reset_index(drop=True)
        df_list.append(df)
    
    return df_list





if __name__ == '__main__':
    import os; os.system('clear')
    from ciervo.io.load import example_marcha_larga
    from ciervo.plots import emg_plot

    data = example_marcha_larga()[0]
    data = label_data(data, num_fases=3)
    emg_plot(data[0])
    


from setuptools import setup, find_packages
import os


VERSION = os.getenv('VERSION')
DESCRIPTION = 'Ciervo for prosthetic leg control'
LONG_DESCRIPTION = 'First version'

# Setting up
setup(
       # the name must match the folder name 'verysimplemodule'
        name="ciervo", 
        version=VERSION,
        author="Carlos Valle",
        author_email="cgvallea@gmail.com",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        packages=find_packages(),
        install_requires=[], # add any additional packages that 
        # needs to be installed along with your package. Eg: 'caer'
        
        keywords=['python', 'first package'],
        classifiers= [
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Education",
            "Programming Language :: Python :: 3",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
            # Linux
            "Operating System :: POSIX :: Linux",
        ],
        package_data={'ciervo': ['tests/data/marcha/*.csv', 'tests/data/marcha_larga/*.csv']},
        include_package_data=True,


)


import os; os.system('clear')
from paho.mqtt import client as mqtt_client
from ciervo.models import features_v1
import numpy as np
import ciervo.parameters as p
from ciervo.aux_tools import Buffer
import time
import joblib
import serial
import struct


broker = p.BROKER_HOST
port = 1883
topic = "data"


class SendAngleSerial:

    def __init__(self):
        self.ser = serial.Serial('/dev/ttyACM0',
                                 baudrate=9600,
                                 bytesize=serial.EIGHTBITS,
                                 parity=serial.PARITY_NONE,
                                 stopbits=serial.STOPBITS_ONE,
                                 timeout=1)
        time.sleep(1)

    def send_float_via_serial(self, value):
        # Send int
        int_value = int(value)
        if 0 <= int_value <= 255:
            self.ser.write(bytes([int_value]))
        else:
            if int_value > 255:
                self.ser.write(bytes([255]))
            
            elif int_value < 0:
                self.ser.write(bytes([0]))



if __name__ == '__main__':

    send = SendAngleSerial()
    print("Enviado")
    send.send_float_via_serial(90)